
// F. Gelis -- Last modified on 2019/06/06



#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "collision.h"


///////////////////////////////////////////////////////////////////////////////////////

double *w;

void init_w(dist_t *F) {
  int i;
  int N = F->N;

  w = (double *)malloc((N+1)*sizeof(double));

  w[0] = 0.5;
  w[N] = 0.5;
  for (i=1;i<N;i++) w[i] = 1.0;
}


///////////////////////////////////////////////////////////////////////////////////////


double p_to_om(double p,double m){
  return sqrt(p*p+m*m);
}


double om_to_p(double om,double m){
  return sqrt((om-m)*(om+m));
}


///////////////////////////////////////////////////////////////////////////////////////


double E_F(dist_t *F){
  double sum = 0.0;
  int N      = F->N;
  int i;

  for (i=0;i<=N;i++) sum += w[i]*F->p[i]*F->om[i]*F->f[i] * F->om[i];

  return F->dom*sum/(2.0*M_PI*M_PI)+(F->n0)*(F->mass);
}


double N_F(dist_t *F){
  double sum = 0.0;
  int N      = F->N;
  int i;

  for (i=0;i<=N;i++) sum += w[i]*F->p[i]*F->om[i]*F->f[i];

  return F->dom*sum/(2.0*M_PI*M_PI)+(F->n0);
}


double Si(double f){ 
  return (1.0+f)*log(1.0+f)-f*log(f);
}


double S_F(dist_t *F){
  double sum = 0.0;
  int N      = F->N;
  int i;

  for (i=0;i<=N;i++) sum += w[i]*F->p[i]*F->om[i] * Si(F->f[i]);

  return F->dom*sum/(2.0*M_PI*M_PI);
}


double Sci(double f){ 
  return log(0.5+f)-log(0.5);
}


double Sc_F(dist_t *F){
  double sum = 0.0;
  int N      = F->N;
  int i;

  for (i=0;i<=N;i++) sum += w[i]*F->p[i]*F->om[i] * Sci(F->f[i]);

  return F->dom*sum/(2.0*M_PI*M_PI);
}


///////////////////////////////////////////////////////////////////////////


double I_full(double f1,double f2,double f3,double f4,double lambda __attribute__((unused))){
  return (1.0+f1+f2)*f3*f4 - f1*f2*(1.0+f3+f4);
}


double I1_full(double f2,double f3,double f4,double n0,double lambda __attribute__((unused))){
  return n0*(f3*f4 - f2*(1.0+f3+f4));
}


double I2_full(double f1,double f3,double f4,double n0,double lambda __attribute__((unused))){
  return n0*(f3*f4 - f1*(1.0+f3+f4));
}


double I3_full(double f1,double f2,double f4,double n0,double lambda __attribute__((unused))){
  return n0*((1.0+f1+f2)*f4 - f1*f2);
}


///////////////////////////////////////////////////////////////////////////


double I_interp(double f1,double f2,double f3,double f4,double lambda){
  return (lambda+f1+f2)*f3*f4 - f1*f2*(lambda+f3+f4);
}


double I1_interp(double f2,double f3,double f4,double n0,double lambda){
  return n0*(f3*f4 - f2*(lambda+f3+f4));
}


double I2_interp(double f1,double f3,double f4,double n0,double lambda){
  return n0*(f3*f4 - f1*(lambda+f3+f4));
}


double I3_interp(double f1,double f2,double f4,double n0,double lambda){
  return n0*((lambda+f1+f2)*f4 - f1*f2);
}


//////////////////////////////////////////////////////////////////////////////


double I_class(double f1,double f2,double f3,double f4,double lambda __attribute__((unused))){
  return (0.5+f3)*(0.5+f4)*(1.0+f1+f2)-(0.5+f1)*(0.5+f2)*(1.0+f3+f4);
}


double I1_class(double f2,double f3,double f4,double n0,double lambda __attribute__((unused))){
  return n0*((0.5+f3)*(0.5+f4)-(0.5+f2)*(1.0+f3+f4));
}


double I2_class(double f1,double f3,double f4,double n0,double lambda __attribute__((unused))){
  return n0*((0.5+f3)*(0.5+f4)-(0.5+f1)*(1.0+f3+f4));
}


double I3_class(double f1,double f2,double f4,double n0,double lambda __attribute__((unused))){
  return n0*((0.5+f4)*(1.0+f1+f2)-(0.5+f1)*(0.5+f2));
}


//////////////////////////////////////////////////////////////////////////////


double I_class0(double f1,double f2,double f3,double f4,double lambda __attribute__((unused))){
  return f3*f4*(f1+f2) - f1*f2*(f3+f4);
}


double I1_class0(double f2,double f3,double f4,double n0,double lambda __attribute__((unused))){
  return n0*(f3*f4 - f2*(f3+f4));
}


double I2_class0(double f1,double f3,double f4,double n0,double lambda __attribute__((unused))){
  return n0*(f3*f4 - f1*(f3+f4));
}


double I3_class0(double f1,double f2,double f4,double n0,double lambda __attribute__((unused))){
  return n0*(f4*(f1+f2) - f1*f2);
}


//////////////////////////////////////////////////////////////////////////////

//#define __M2__(s,t,u) (10.0/(1.0+1.0*((s)*(s)+(t)*(t)+(u)*(u))))
//#define __M2__(s,t,u) (0.0)
//#define __M2__(s,t,u) ((s*s+t*t+u*u > 0.001) ? 10.0 : 0.0)

inline double __M2__phi3(double s,double t,double u){
        double num = s*t+t*u+u*s;
        double den = s*(t-0.1)*(u-0.1);
        
        return num*num/(den*den);
}


inline double __M2__(double s,double t,double u){
        return 3.0 - t*u/(s*s) - u*s/((t-0.1)*(t-0.1)) - s*t/((u-0.1)*(u-0.1));
}


inline double __M2__phi4(double s,double t,double u){
        return 5.0/(1.0+0.01*(s*s+t*t+u*u));
}

double M2(double s0,double t0,double u0,double al){ // assumes Mpsi multiple of 4, and (t,u) symmetry
  int n;
  double Psi,dPsi = 2.0*M_PI/MPsi;
  double sum = 0.0;
  double t,u;

  // n = 0 and MPsi/2 (cosine = 1 and -1)
  t    = t0 + al;
  u    = u0 - al;
  sum += 2.0 * __M2__(s0,t,u);

  // n = MPsi/4 and 3*MPsi/4 (cosine = 0)
  t    = t0;
  u    = u0;
  sum += 2.0 * __M2__(s0,t,u);

  // Other values: 1 to MPsi/4-1
  for (n=1;n<MPsi/4;n++){
    Psi  = n*dPsi;
    t    = t0 + al*cos(Psi);
    u    = u0 - al*cos(Psi);
    sum += 4.0 * __M2__(s0,t,u);
  }
  
  return sum*dPsi;
}


double M2_2(double s0,double t0,double u0,double al){ // assumes Mpsi multiple of 2
  int n;
  double Psi,dPsi = 2.0*M_PI/MPsi;
  double sum = 0.0;
  double t,u;

  // n = 0  (cosine = 1)
  t    = t0 + al;
  u    = u0 - al;
  sum += __M2__(s0,t,u);

  // n = MPi/2  (cosine = -1)
  t    = t0 - al;
  u    = u0 + al;
  sum += __M2__(s0,t,u);

  // Other values: 1 to MPsi/2-1
  for (n=1;n<MPsi/2;n++){
    Psi  = n*dPsi;
    t    = t0 + al*cos(Psi);
    u    = u0 - al*cos(Psi);
    sum += 2.0 * __M2__(s0,t,u);
  }
  
  return sum*dPsi;
}

//////////////////////////////////////////////////////////////////////////////


inline double MatEl2_(double d12,double d34,double s12,double s34,\
	       double om1, double om2,double om3,double om4,	   \
	       double P,dist_t *F){
  double m2   = F->mass2;
  double invP = 1.0/P;
  double a12  = d12*invP;
  double a34  = d34*invP;
  
  double Nc1  = P+s12*a12;
  double Nc3  = P+s34*a34;
  double Nc4  = P-s34*a34;
  double e1   = (s12+P)*(s12-P);
  double e2   = (s34+P)*(s34-P);
  double e3   = 1.0-a12*a12; // (1.0+a12)*(1.0-a12);
  double e4   = 1.0-a34*a34; // (1.0+a34)*(1.0-a34);
  double Ns13 = sqrt(e1*e2*e3*e4);
  double om12 = om1+om2;
  double s0   = (om12+P)*(om12-P);
  double t0   = 2.0*(m2-om1*om3)+0.5*Nc1*Nc3;
  double u0   = 2.0*(m2-om1*om4)+0.5*Nc1*Nc4;
  double al   = 0.5*Ns13;

  //fprintf(stderr,"(s+t+u-4m2)/4m2 = %.5e\n",(s+t+u-4.0*m2)/(4.0*m2));
    
  return M2(s0,t0,u0,al);
}


double MatEl2_f(double p1,double p2,double p3,double p4,\
		double om1,double om2,double om3,double om4,\
		double pmin,dist_t *F){
  int m;
  double P0;
  double P;
  double d12 = p1-p2;
  double d34 = p3-p4;
  double ab12 = fabs(d12);
  double ab34 = fabs(d34);
  double dP   = pmin / MP;
  double sum  = 0.0;
  
  P0 = ((ab12>ab34) ? ab12 : ab34) + 0.5*dP;
  
  for (m=0;m<MP;m++){
    P = P0 + m*dP;
    sum += MatEl2_(d12,d34,p1+p2,p3+p4,om1,om2,om3,om4,P,F);
  }
  
  return sum*dP;
}

double MatEl2_n1(double p1,double p2,double p3,double p4,double om1,double om2,double om3,double om4,double pmin __attribute__((unused)),dist_t *F){

  return MatEl2_(p1-p2,p3-p4,p1+p2,p3+p4,om1,om2,om3,om4,p2,F);
}

double MatEl2_n2(double p1,double p2,double p3,double p4,double om1,double om2,double om3,double om4,double pmin __attribute__((unused)),dist_t *F){

  return MatEl2_(p1-p2,p3-p4,p1+p2,p3+p4,om1,om2,om3,om4,p1,F);
}

double MatEl2_n3(double p1,double p2,double p3,double p4,double om1,double om2,double om3,double om4,double pmin __attribute__((unused)),dist_t *F){

  return MatEl2_(p1-p2,p3-p4,p1+p2,p3+p4,om1,om2,om3,om4,p4,F);
}


///////////////////////////////////////////////////////////////////////////////

#define w0 0.5
#define wN 0.5
#define wi 1.0


double coll_f(int i,int k,int j,double(*II)(double,double,double,double,double),dist_t *F,int min){
  double f1   = F->f[i];
  double f3   = F->f[j];
  double f4   = F->f[k];
  double f2   = F->f[j+k-i];
  double pmin = F->p[min];
  double p1   = F->p[i];
  double p3   = F->p[j];
  double p4   = F->p[k];
  double p2   = F->p[j+k-i];
  double om1  = F->om[i];
  double om3  = F->om[j];
  double om4  = F->om[k];
  double om2  = F->om[j+k-i];

  return pmin*II(f1,f2,f3,f4,F->lambda)*MatEl2_f(p1,p2,p3,p4,om1,om2,om3,om4,pmin,F);
}


double coll_term_f(int i,double(*II)(double,double,double,double,double),dist_t *F,double g){
  double sum = 0.0;
  double p1  = F->p[i];
  double om1 = F->om[i];
  int N      = F->N;
  int j,k;

  // min k (p3)

  for (j=i+1;j<=N;j++) {
    for (k=0;k<i;k++){
      sum += w[j]*w[k]*w[j+k-i]*coll_f(i,j,k,II,F,k);
    }
  }
 
  // min j (p4)

  for (j=0;j<i;j++) {
    for (k=i+1;k<=N;k++){
      sum += w[j]*w[k]*w[j+k-i]*coll_f(i,j,k,II,F,j);
    }
  }

  // min p1

  for (j=i+1;j<=N;j++) {
    for (k=i+1;k<=N+i-j;k++) {
      sum += w[j]*w[k]*w[j+k-i]*coll_f(i,j,k,II,F,i);
    }
  }

  // min p2

  for (j=0;j<i;j++) {
    for (k=i-j;k<i;k++) {
      sum += w[j]*w[k]*w[j+k-i]*coll_f(i,j,k,II,F,j+k-i);
    }
  }

  sum += w[i]*w[i]*w[i]*coll_f(i,i,i,II,F,i);

  for (j=0;j<i;j++)    sum += w[j]*w[i]*w[j]*coll_f(i,j,i,II,F,j);
  for (j=i+1;j<=N;j++) sum += w[j]*w[i]*w[j]*coll_f(i,j,i,II,F,i);
  for (k=0;k<i;k++)    sum += w[i]*w[k]*w[k]*coll_f(i,i,k,II,F,k);
  for (k=i+1;k<=N;k++) sum += w[i]*w[k]*w[k]*coll_f(i,i,k,II,F,i);


  return sum*(F->dom*F->dom)*pow(g,4.0)/(p1*om1*128.0*M_PI*M_PI*M_PI);
}


////////////////////////////////////////////////////////////////////////////////


double coll_n1(int k,int j,double(*I1)(double,double,double,double,double),dist_t *F){
  double f3   = F->f[j];
  double f4   = F->f[k];
  double f2   = F->f[j+k+1];
  double pmin = 0.0;
  double p1   = 0.0;
  double p3   = F->p[j];
  double p4   = F->p[k];
  double p2   = F->p[j+k+1];
  double om1  = F->mass;
  double om3  = F->om[j];
  double om4  = F->om[k];
  double om2  = F->om[j+k+1];

  return I1(f2,f3,f4,F->n0,F->lambda)*MatEl2_n1(p1,p2,p3,p4,om1,om2,om3,om4,pmin,F);

}


double coll_term_n01(double(*I1)(double,double,double,double,double),dist_t *F,double g){
  double sum = 0.0;
  int N      = F->N;
  int j,k;

  for (k=0;k<=N-1;k++){
    for (j=0;j<=N-k-1;j++){
      sum += w[k]*w[j]*w[j+k+1]*coll_n1(j,k,I1,F);
    }
  }

  return sum*(F->dom*F->dom)*pow(g,4.0)/(F->mass*128.0*M_PI*M_PI*M_PI);
}


///////////////////////////////////////////////////////////////////////////////////


double coll_n2(int i,int k,int j,double(*I2)(double,double,double,double,double),dist_t *F){
  double f3   = F->f[j];
  double f4   = F->f[k];
  double f1   = F->f[i];
  double pmin = 0.0;
  double p1   = F->p[i];
  double p3   = F->p[j];
  double p4   = F->p[k];
  double p2   = 0.0;
  double om1  = F->om[i];
  double om3  = F->om[j];
  double om4  = F->om[k];
  double om2  = F->mass;

  return I2(f1,f3,f4,F->n0,F->lambda)*MatEl2_n2(p1,p2,p3,p4,om1,om2,om3,om4,pmin,F);

}


double coll_term_n02(int i,double(*I2)(double,double,double,double,double),dist_t *F,double g){
  double sum = 0.0;
  double p1  = F->p[i];
  double om1 = F->om[i];
  int j;

  for (j=0;j<=i-1;j++) sum += w[i-j-1]*w[j]*coll_n2(i,i-j-1,j,I2,F);

  return sum*(F->dom)*pow(g,4.0)/(F->mass*p1*om1*64.0*M_PI);
}


///////////////////////////////////////////////////////////////////////////////////


double coll_n3(int i,int k,int j,double(*I3)(double,double,double,double,double),dist_t *F){
  double f4   = F->f[j];
  double f2   = F->f[k];
  double f1   = F->f[i];
  double pmin = 0.0;
  double p1   = F->p[i];
  double p3   = 0.0;
  double p4   = F->p[j];
  double p2   = F->p[k];
  double om1  = F->om[i];
  double om3  = F->mass;
  double om4  = F->om[j];
  double om2  = F->om[k];

  return I3(f1,f2,f4,F->n0,F->lambda)*MatEl2_n3(p1,p2,p3,p4,om1,om2,om3,om4,pmin,F);

}


double coll_term_n03(int i,double(*I3)(double,double,double,double,double),dist_t *F,double g){
  double sum = 0.0;
  double p1  = F->p[i];
  double om1 = F->om[i];
  int N      = F->N;
  int j;

  for (j=i+1;j<=N;j++) sum += w[j-i-1]*w[j]*coll_n3(i,j-i-1,j,I3,F);

  return sum*(F->dom)*pow(g,4.0)/(F->mass*p1*om1*32.0*M_PI);
}


///////////////////////////////////////////////////////////////////////////////////


double feq_full(double p){
  return 1.0/(exp(sqrt(p*p+0.25))-1.0);
}


double feq_class(double p){
  return 6.0/sqrt(p*p+0.25)-0.5;
}


double f1(double p){
  return 10.0/(exp(p)*exp(p)+1.0);
}


double f2(double p){
  return (p<1.0) ? 5.0 : 0.0;
}


double f3(double p){
  return 2.0/(1.0+exp((p-1.0)/0.1));
}


double f4(double p){
  return 6.0/(exp(sqrt(p*p+0.25))-1.0);
}


double naoto(double p){
  return 5.0*exp(-p*p/9.0);
}


double cgc(double p){
  return 6.0*((p<1.0) ? 1.0 : 0.0);
}


dist_t *F_init_om(int N,double mass,double pmax,double(*f)(double),double lambda){
  dist_t *F=(dist_t *)malloc(sizeof(dist_t));
  int i;
  double omMax = p_to_om(pmax,mass);
  double dom   = (omMax-mass)/((double)(N+1));
  double omMin = mass+dom;
  double om;

  F->f          = (double*)malloc((N+1)*sizeof(double));
  F->p          = (double*)malloc((N+1)*sizeof(double));
  F->om         = (double*)malloc((N+1)*sizeof(double));
  F->N          = N;
  F->mass       = mass;
  F->mass2      = mass*mass;
  F->lambda     = lambda;
  F->pmin       = om_to_p(omMin,mass);
  F->pmax       = pmax;
  F->omMin      = omMin;
  F->omMax      = omMax;
  F->dom        = dom;
  F->invdom     = 1.0/dom;
  F->n0         = 1.0e-5;
  
  for (i=0;i<=N;i++){
    om = omMin + i*dom;
    F->om[i]   = om;
    F->p[i]    = om_to_p(om,mass);
    F->f[i]    = f(F->p[i]);
  }

  return F;
}


/////////////////////////////////////////////////////////////////////////////////////////


int increment_F_Euler(double dt,double g,dist_t *F,quant_t approx){
  int i,status;
  double *df  = NULL;
  double dn0  = 0.0;
  int N1 = 1+F->N;
  df  = (double*)malloc((F->N+1)*sizeof(double));
  double (* I)(double,double,double,double,double);
  double (*I1)(double,double,double,double,double);
  double (*I2)(double,double,double,double,double);
  double (*I3)(double,double,double,double,double);

  if (approx==QUANTUM)
    {
      I   = I_full;
      I1  = I1_full;
      I2  = I2_full;
      I3  = I3_full;
    } 
  else if (approx==CLASSICAL) 
    {
      I   = I_class;
      I1  = I1_class;
      I2  = I2_class;
      I3  = I3_class;
    } 
  else if (approx==CLASSICAL0)
    {
      I   = I_class0;
      I1  = I1_class0;
      I2  = I2_class0;
      I3  = I3_class0;
    }
  else
    {
      I   = I_interp;
      I1  = I1_interp;
      I2  = I2_interp;
      I3  = I3_interp;
    }
  
  /////// condensate --> quick
  dn0 = dt*coll_term_n01(I1,F,g);
  /////////////////////////

    ///////// Bulk: main loop ---> very slow
#pragma omp parallel for num_threads(NT)
  for (i=0;i<N1;i++){
    df[i] = dt*(coll_term_f  (i,I, F,g) + \
		coll_term_n02(i,I2,F,g) + \
		coll_term_n03(i,I3,F,g));
  }
  ///////////////////////////////////////
  

  // check if the new f[i] has negative values
  status=0;
  for (i=0;i<=F->N;i++) {if ((F->f[i]+df[i])<0) status = 1;}
  
  // If the new f[i] has negative values, we return immediatel:
  // the timestep must be recomputed
  if (1==status) return 1;

  // Otherwise, update f[i] and n0
  for (i=0;i<=F->N;i++) F->f[i] += df[i];
  F->n0 += dn0;
  free(df);
  
  
  return 0;
}
