
//  F. Gelis -- last modified on 2019/05/20


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <mpi.h>

#include "collision.h"

#define min(x,y) (0.5*((x)+(y)-fabs((x)-(y))))


int main(int argc,char *argv[]){
  int N       = NP;
  double Q    = 1.0;
  ///////////////////////////////////////////////////////////
  quant_t evol_type = QUANTUM;
  double lambda = 1.0;
  double mass   = 0.1*Q;
  double pmax   = 6.0*Q;
  double g      = 1.0;
  dist_t *F     = F_init_om(N,mass,pmax,cgc,lambda);
  ///////////////////////////////////////////////////////////
  double dt0      = 1.0/Q;
  double t        = 0.0/Q;
  double dnc_rel  = 0.01;
  double dnc_abs  = 0.001;
  double freq     = 5.0;
  ///////////////////////////////////////////////////////////
  double dt;
  double E,n,S,Sc,E0,n0;
  int i;
  char *filename = "out.txt";
  char *logfile  = "log.txt";
  char *monfile  = "mon.txt";
  FILE *FL = NULL;
  FILE *FF = NULL;
  FILE *FM = NULL;
  time_t tm0;
  char * tm1;
  double nc0,nc1;
  int lastsaved;
  int status;

  int rank;
  int size;

  MPI_Init(&argc,&argv);               // initialize MPI
  MPI_Comm_size(MPI_COMM_WORLD,&size); // number of CPUs
  MPI_Comm_rank(MPI_COMM_WORLD,&rank); // rank in the pool of CPUs

  init_w(F);

  E0 = E_F(F);
  n0 = N_F(F);
  S  = S_F(F);
  Sc = Sc_F(F);
  

  if (rank==0){
    FF = fopen(filename,"w");
    for (i=0;i<=F->N;i++){
      fprintf(FF,"%.4f %.6e %.6f %.10f\n",t,F->p[i],F->om[i],F->f[i]);
    }
    lastsaved=0;
    fprintf(FF,"\n");
    fclose(FF);

    FL = fopen(logfile,"w");
    tm0 = time(NULL);
    tm1 = strtok(ctime(&tm0), "\n");
    fprintf(FL,"%s : t = %.6f\n",tm1,t);
    fclose(FL);

    FM = fopen(monfile,"w");
    fprintf(FM,"%.6f %.6e %.10f %.10f %.10f %.10f %.4e %.4e\n",t,F->n0,E0,n0,S,Sc,0.0,0.0);
    fclose(FM);

  }

  dt  = dt0;
  nc0 = F->n0;

  while (t<=200.0){

    MPI_Barrier(MPI_COMM_WORLD);

    status = 1;
    while (1==status) {
      status = increment_F_Euler(dt,g,F,evol_type,size,rank);
      if (1==status) {
	dt = dt/4.0;
	if (rank==0) fprintf(stderr,"t = %.3f : dt -> dt/4 because f<0 : dt = %.6f\n",t,dt);
      }
    }

    t += dt;

    MPI_Barrier(MPI_COMM_WORLD);

    nc1 = F->n0;
    
    if ((fabs((nc1-nc0)/nc1)>dnc_rel)||(fabs(nc1-nc0)>dnc_abs)) dt = dt*0.5;
    if ((fabs((nc1-nc0)/nc1)<dnc_rel)&&(fabs(nc1-nc0)<dnc_abs)) dt = min(dt*1.1,dt0);
    nc0 = nc1;
    
    if (rank==0){
      if  (((t/freq-floor(t/freq)) < dt0/freq)&&(((int)floor(t))!=lastsaved)){
	lastsaved = (int)floor(t);
	FF = fopen(filename,"a");
	for (i=0;i<=F->N;i++){
	  fprintf(FF,"%.4f %.6e %.6f %.10f\n",t,F->p[i],F->om[i],F->f[i]);
	}
	fprintf(FF,"\n");
	fclose(FF);
      }
      
      E = E_F(F);
      n = N_F(F);
      S  = S_F(F);
      Sc = Sc_F(F);
      
      FL=fopen(logfile,"a");
      tm0   = time(NULL);
      tm1 = strtok(ctime(&tm0), "\n");
      fprintf(FL,"%s : t = %.6f\n",tm1,t);
      fclose(FL);

      FM = fopen(monfile,"a");
      fprintf(FM,"%.6f %.6e %.10f %.10f %.10f %.10f %.4e %.4e\n",\
	      t,F->n0,E,n,S,Sc,(E-E0)/E0,(n-n0)/n0);
      fclose(FM);
    }

  }

  MPI_Finalize();
  exit(0);
}
